
xVersion" 2012.10.10.1"

module Admin


  #document Admin

    // Admin.4th
    //
    //   Settings in persistant storage

  #endDoc


private

  uses
    hw_interface                          // For KickDog in AdminCalcCrc
    Flash
    utils                                 // for CRC stuff
    strings                               // For string access handlers
  endUses


    1000.000 2constant FlashMaxCycles     #// Maximum number of erase / write cycles for Flash device


  // //////////////////////////////////////////////////////////////////////////
  //
  // Meta helper code for Flash management
  //
  {
    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    // Eprom sizes

    512 1024 * constant Flash1Size          #// Total size in bytes of Flash 2

    256        constant FlashPageSize       #// Page size for both Flashs

    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    variable      AdminHere                 #// maintain a Here variable for offset and size calculations
    varInterface  AdminHere

    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : AdminAlloc ( n -- )

      #// Allocate space in an Flash

      AdminHere +!
    ;


    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : AdminLabel ( -- ) // < token -- >

      #// Set a label at the current location in an Flash

      AdminHere@ 2constant
    ;


    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : AdminField ( nFieldSize -- ) // < token -- >

      #// Define a label at the current Flash location and reserve space

      AdminLabel
      AdminAlloc
    ;


    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : AdminEntry ( nSize -- ) // < token -- >

      #// Create an admin entry by defining some constants

      AdminHere@
      bl token dup                    // Read next token from input and dup it
      " Size" +                       // Append Size
      bl untoken                      //   and push back into input
      bl untoken                      // Also push back unaltered version
      2constant                       // Offset is a dword
      dup
      constant                        // Size is a word
      AdminAlloc
    ;


    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : AdminPageAlign ( -- )

      #// Start at an Flash page boundary

      AdminHere@
      FlashPageSize 1 - +
      FlashPageSize /
      FlashPageSize *
      AdminHere!
    ;


  }
  //
  // End of meta code


  // Target support stuff

  256 constant AdminCrcBlockSize      #// Block size for CRC calculations


  AdminCrcBlockSize
    array   AdminBuffer               #// an Flash read / write buffer

  2variable AdminOffset               #// Used for CRC and for boot loader
   variable AdminCount                #// Used for CRC and for boot loader
  2variable AdminIndex                #// Used for CRC and for boot loader
   variable AdminSeed                 #// Used for boot loader

  dvarInterface AdminOffset
   varInterface AdminCount
  dvarInterface AdminIndex
   varInterface AdminSeed


  : @AdminBuffer      (   -- pRam ) AdminBuffer ;
  : AdminBuffer@      (   -- b    ) @AdminBuffer  @ ;
  : AdminBuffer!      ( b --      ) @AdminBuffer  ! ;
  : AdminBufferDup!   ( b -- pRam ) @AdminBuffer  ! @AdminBuffer ;
  : AdminBufferDupd!  ( w -- pRam ) @AdminBuffer d! @AdminBuffer ;
  : AdminBufferd@     (   -- w    ) @AdminBuffer d@ ;


  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  //
  // Flash 1 layout - mapped over Flash 2 (both are the same physical device)
  //
  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////


  // Item sizes

  // //////////////////////////////////////////////////////////////////////////
  //
  // Space reservations / offset definitions

  $100 2constant AdminReservedLS                    #// Number of bytes reserved for Loader and Serial number
                                                    #// All data before this marker will be persistent
                                                    #// and is reserved for boot loader runtime management
                                                    #// and for a serial number.

  $3000 2constant BootLoaderImageStart              #// Start of flash image in Flash 2
                                                    #// Symbol BootLoaderImageStart also used in  18fxxj11.
                                                    #//
                                                    #//   !! BootLoaderImageStart Mod 256 MUST BE ZERO !!
                                                    #//
                                                    #// All data before this address is persistant, the area
                                                    #// from AdminReservedLS to BootLoaderImageStart is meant
                                                    #// for user configuration data that must survive a reflash
                                                    #// operation.


  // Reserve AdminReservedLS bytes for serial number and bootloader management

  AdminReservedLS AdminHere!                        #// Start at offset in Flash
                                                    #// to keep loader and serial nr. info
                                                    #// accessible always.

  // //////////////////////////////////////////////////////////////////////////
  // User configuration space

  AdminLabel  AdminUserConfigStart                  #// Start offset of user configuration data

  // //////////////////////////////////////////////////////////////////////////
  // Sanity checks for Flash1

  AdminHere@ constant Admin1TotalSize             #// Total area in use in Flash (in bytes)

  Admin1TotalSize Flash1Size >
  #if
    runerror' data structure too large for Flash 1 device'
  #endif

  // end sanity checks for Flash1
  // //////////////////////////////////////////////////////////////////////////


  // //////////////////////////////////////////////////////////////////////////
  // Boot loader stuff

  0 AdminHere!                                     // Start at offset 0 in Flash

  $0000 constant b_loader_state           $01 constant b_loader_state_size
  $0001 constant b_loader_state_check     $01 constant b_loader_state_check_size
  $0002 constant b_image_size             $02 constant b_image_size_size
  $0004 constant b_checksum               $01 constant b_checksum_size

  $0005 constant b_size                           #// Total size of fields above


  // //////////////////////////////////////////////////////////////////////////
  // Checksum stuff

  : AdminIndex++x ( -- )

    AdminIndex@
    AdminCrcBlockSize 0 d+
    AdminIndex!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminCount--x ( -- f )

    #// Increment index, decrement counter, return true when zero

    AdminIndex++x
    AdminCount@
    AdminCrcBlockSize
    >
    if
      False             // Return value
      AdminCount@       // New AdminCount value
      AdminCrcBlockSize
      -
    else
      True              // Return value
      0                 // New AdminCount value
    endIf
    AdminCount!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminCrcSetup ( dStartOffset nSize -- )

    AdminCount!
    AdminOffset!
    d0 AdminIndex!
    0 crc!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminGetByteCount ( -- n )

    AdminCount@
    AdminCrcBlockSize
    >
    if
      AdminCrcBlockSize
    else
      AdminCount@
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminChkBytes2 ( -- )

    AdminIndex@                     // Get current offset
    AdminOffset@ d+                 // and make double
    AdminGetByteCount               // ( wCount )
    @AdminBuffer                    // Ram   address
    Flash1ReadBlockRam

    AdminGetByteCount               // ( bCount )
    @AdminBuffer
    BlockCrc16Nc                    // wants ( bCount pRam -- )
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminCalcCrc2 ( wStartAddr wSize -- wCrc )

    AdminCrcSetup   // ( )
    begin
      AdminChkBytes2
      KickDog
      AdminCount--x
    until
    Crc@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : AdminCheckCrc2 ( wStartAddr wSize -- f )

    dup 0=
    if
      ddrop false
    else
      AdminCalcCrc2
      0=
    endif
  ;


  // //////////////////////////////////////////////////////////////////////////
  // Boot loader support

  //  $0000 2constant b_loader_state          $01 constant b_loader_state_size
  //  $0001 2constant b_loader_state_check    $01 constant b_loader_state_check_size
  //  $0002 2constant b_image_size            $02 constant b_image_size_size
  //  $0004 2constant b_checksum              $02 constant b_checksum_size


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderFixCrc ( -- )

    b_loader_state 4.
    AdminCalcCrc2                   // ( wStartAddr wSize -- wCrc )
    AdminBufferDupd!
    b_checksum
    b_checksum_size
    Flash1WriteBlockPaged
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderControl@ ( -- w ) #// Get loader_state

    b_loader_state_check 0          // E2rom address
    b_loader_state_size             // wByteCount
    @AdminBuffer                    // Ram   address
    Flash1ReadBlockRam
    AdminBuffer@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderState@ ( -- w ) #// Get loader_state

    b_loader_state 0                // E2rom address
    b_loader_state_size             // wByteCount
    @AdminBuffer                    // Ram   address
    Flash1ReadBlockRam
    AdminBuffer@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderState! ( w -- ) #// Set loader_state and loader_state_check

    dup not                         // ( state !state ) treated as a dword entity
    AdminBufferDupd!                // Write into AdminBuffer
    b_loader_state                  // E2rom address
    4                               // wByteCount
    Flash1WriteBlockPaged
    BootLoaderFixCrc                // Fix CRC
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderCount@ ( -- d ) #// Get loader byte count

    b_image_size 0                  // E2rom address
    b_image_size_size               // wByteCount
    @AdminBuffer                    // Ram   address
    Flash1ReadBlockRam
    AdminBufferd@
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderCount! ( d -- ) #// Set loader byte count

    AdminBufferDupd!                // Write into AdminBuffer
    b_image_size                    // E2rom address
    b_image_size_size               // wByteCount
    Flash1WriteBlockPaged
    BootLoaderFixCrc                // Fix CRC
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderStateOk? ( -- f ) #// Get checksum and state validity

    b_loader_state 6.               // Check 6 bytes, crc must be 0
    AdminCheckCrc2
    dup
    if
      BootLoaderState@
      BootLoaderControl@ Not =
      And
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderFixState ( -- ) #// Fix invalid bootloader state or checksum

    BootLoaderControl@ Not
    BootLoaderState@ =
    b_loader_state 6.               // Check 6 bytes, crc must be 0
    AdminCheckCrc2
    And
    if
      BootLoaderState@
      case
        0    of False endOf         // All Ok, apply no fix
        1    of False endOf         // All Ok, apply no fix
        2    of False endOf         // All Ok, apply no fix
        3    of False endOf         // All Ok, apply no fix
        $ff  of False endOf         // All Ok, apply no fix
      default
        Drop True                   // Wrong state, apply fix
      endCase
    else
      True                          // Error in CRC or error in control byte, apply fix
    endIf

    if
      $ff BootLoaderState!          // Fix boot loader state into something harmless
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderStart ( -- d ) #// First boot loader programmable location in Flash 2

    BootLoaderImageStart
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderWriteData ( $ram -- f ) #// Write a block of loader data

    dup >r >r                       // Count twice on loop stack ( pRam )
    BootLoaderStart
    BootLoaderCount@ d+             // ( pRam dFlashAddr )
    r>                              // ( pRam dFlashAddr nCount )
    Flash1WriteBlock                //
    r> 0                            // Get 2nd copy nCount as a dword
    BootLoaderCount@ +              // and add it to the offset
    BootLoaderCount!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderReadData ( $ram -- $ram )

    #// Read a block of loader data

  ;

  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderCheckImage ( -- f ) #// Check image CRC to be zero

    BootLoaderStart
    BootLoaderCount@
    AdminCheckCrc2
  ;


  // //////////////////////////////////////////////////////////////////////////

  : blsize  ( -- b ) 64 ;

  : AdminBufferia+ ( -- pRam ) AdminBuffer i a+ ;

  : BootLoaderDecrypt ( -- )

    43 AdminSeed!
    BootLoaderCount@
    61. d+                  // Minus 2 rounded to multiple of block size
    blsize /                //   ...
    blsize *                //   ...
    AdminCount!             //   stored as AdminCount
    BootLoaderStart
    AdminOffset!
    begin
      AdminCount@ 0<>
    while
      KickDog
      AdminOffset@          // Flash address
      blsize                // byte count
      @AdminBuffer          // Ram    address
      Flash1ReadBlockRam
      blsize 0
      do
        AdminBufferia+ @
        AdminSeed@
        Xor Dup
        AdminBufferia+ !
        Random AdminSeed!
      loop
      @AdminBuffer          // Ram    address
      AdminOffset@          // Flash address
      blsize                // byte count
      Flash1WriteBlockPaged
      AdminCount@  blsize    - AdminCount!
      AdminOffset@ blsize 0 d+ AdminOffset!
    repeat
  ;


  // //////////////////////////////////////////////////////////////////////////

  : BootLoaderTrashImage ( -- ) #// Erase loader image

    #// note : this erases user data in Flash
    #// Erase Flash 2 locations $0100 .. $ffff
    #// partially to $ff, partially to whatever is in ram
    #// When done re-initialize the log list to an empty list

    AdminCrcBlockSize 0
    do
      $ff AdminBufferia+ !
    loop
    0. BootLoaderCount!
    begin
      BootLoaderCount@ [ $ffffff00 BootLoaderImageStart - dLiteral ] <
    while
      KickDog
      @AdminBuffer
      AdminCrcBlockSize
      BootLoaderWriteData
      Drop
      BootLoaderCount@
      AdminCrcBlockSize 0 d+
      BootLoaderCount!
    repeat
  ;


  // //////////////////////////////////////////////////////////////////////////

public


  export BootLoaderState@     (       -- b    ) // Get loader_state
  export BootLoaderState!     ( b     --      ) // Set loader_state
  export BootLoaderCount@     (       -- w    ) // Get loader byte count
  export BootLoaderCount!     ( w     --      ) // Set loader byte count
  export BootLoaderStateOk?   (       -- f    ) // Get checksum and state validity
  export BootLoaderFixState   (       --      ) // Fix invalid bootloader state or checksum
  export BootLoaderStart      (       -- d    ) // First programmable location
  export BootLoaderWriteData  ( $ram  -- f    ) // Write a block of loader data
  export BootLoaderReadData   ( $ram  -- $ram ) // Read a block of loader data
  export BootLoaderCheckImage (       -- f    ) // Check image CRC
  export BootLoaderDecrypt    (       --      ) // Decrypt the Flash image
  export BootLoaderTrashImage (       --      ) // Trash loader image


endModule


