
xVersion" 2012.10.10.1"

module forthInterpreter

  done' 2012-10-10 : implement a link chain for deferred words'
  done' 2012-10-11 : implement full memory access for c@, c!, w@, w!, @ and ! - 32 bit entities must not cross an EDS page boundary'
  //
  // This allows for Flash access too .. but its messy ... section 45 of fam. ref. :
  //
  // The entire PM can be accessed through EDS pages, 0x200 to 0x3FF.  The EDS
  // pages, 0x200 to 0x2FF, comprise lower words, while the EDS pages, 0x300 to
  // 0x3FF, comprise upper bytes.  The data from the desired PM location can be
  // read by selecting the appropriate page.  Each page can access 32 Kbytes of
  // data.

  todo' implement thunks for 32 bit code to call into 16 bit code'
  todo' implement thunks for 16 bit code to call into 32 bit code'
  //
  // model for 32 bit code (forth):
  //
  //   TOS    w0, w1                 ; // Top Of Stack Cache Register
  //   IP     TBLPAG, w8             ; // Instruction Pointer, walks over code pointer lists.
  //   W      w6, w7                 ; // W Register, used to compute PFA at run time
  //   PSP    w9                     ; // Parameter Stack Pointer - maps onto the compiled forth = grows upwards
  //   RSP    w10                    ; // Return    Stack Ponter  - maps onto the compiled forth - grows downwards
  //   UP     in memory              ; // As a variable
  //
  //
  // model for 16 bit code (4th) (it does not have UP):
  //
  //   .equiv  r_tos, w0             ; // Cached Top Of data Stack
  //
  //   .equiv  tmp1 , w1             ; // Scratch register for compiled forth, high word of cached TOS for internal forth
  //   .equiv  tmp2 , w2             ; // Scratch registers
  //   .equiv  tmp3 , w3
  //   .equiv  tmp4 , w4
  //   .equiv  tmp5 , w5
  //   .equiv  tmp6 , w6             ; // scratch, high word of W register for internal forth engine
  //   .equiv  tmp7 , w7             ; // scratch, low  word of W register for internal forth engine
  //   .equiv  tmp8 , w8             ; // scratch, Low  word of IP register for internal forth, high word is in TBLPAG
  //
  //   .equiv  DSP  , w9             ; // Data stack pointer in both forth and 4th
  //   .equiv  LSP  , w10            ; // Loop stack pointer, return and loop stack pointer for forth
  //   .equiv  IND0 , w11            ; // Indirection register 0, scratch really
  //   .equiv  IND1 , w12            ; // Indirection register 1, scratch really
  //   .equiv  EFP  , w13            ; // word : Exception Frame Pointer,
  //                                 ; //        used for exception handling
  //   .equiv  LDP  , w14            ; // byte : Local Data Pointer,
  //                                 ; //        Base pointer for local variable access
  //                                 ; //        mapped onto the frame pointer
  //   .equiv  RSP  , w15            ; // Return stack pointer, processor managed

  #document forthInterpreter
    //
    // ////////////////////////////////////////////////////////////////////////
    //
    // This is an internal forth engine, with some assembly level magic words
    // can be added to it from the cross compiler, but also when finished it
    // will be possible to feed in forth source code that will then be
    // compiled. That is the plan at least :-)
    //
    // Using direct threading according to the jump model from:
    //
    //   http://www.figuk.plus.com/build/heart.htm

    // also see:
    //
    //   http://www.bradrodriguez.com/papers/moving1.htm
    //
    // The computed jumps are being performed on registers that are pushed onto
    // the processor hardware stack, this when followed by a return instruction
    // effectively performs a computed jump. (GOTO/BRA)
    //
    // Where DATA was used in the first URL I'll use PFA. doVar will differ in
    // that it returns the RAM address that will need be compiled into PFA  by
    // Variable (as RAM is in another program space than FLASH) ... a Harvard
    // architectural quirk ... anyway, a variable is a constant that is used
    // as a pointer into RAM. The @ and ! operator will work on RAM and not on
    // FLASH. Latr on , and F, and such and FALLOT will work in code (FLASH)
    // space. VALLOT works in variable (RAM) space.
    //
    // This is a 32 bit forth using a cell size of four bytes although in FLASH
    // only 24 bits are valid, the upper 8 ones are zero by definition. And
    // for RAM pointers 17 bits are in use, so this should be able to execute
    // from any FLASH location and be able to use any RAM location. The return
    // and data stack pointers are only 16 bits wide, meaning return and data
    // stack will have to reside in the lower 64 k of RAM.
    //
    // Primaries are just machine code that end in "goto <*" t.next">".
    // Secondaries are pointer lists that start like a primary with code
    // handler bra <*" t.doCol">, variables are pointers to RAM that get
    // compiled into flash as a constant and constants are lteral values
    // compiled into FLASH, the value being pushed by bra <*" t.doCon">.
    //
    // Register allocations used :
    //
    //   TOS    w0, w1                              // Top Of Stack Cache Register
    //   IP     TBLPAG, w8                          // Instruction Pointer, walks over code pointer lists.
    //   W      w6, w7                              // W Register, used to compute PFA at run time
    //   PSP    w9                                  // Parameter Stack Pointer - maps onto the compiled forth
    //   RSP    w10                                 // Return    Stack Ponter  - maps onto the compiled forth
    //   UP     in memory                           // As a variable
    //   HERE   in memory                           // As a variable
    //
    //   TOS == Top  Of data/parameter Stack
    //   NOS == Next On data/parameter Stack , the 2nd stack element.
    //
    // Header layout used for compiled forth words
    //
    //  LFA : + 0 :: Link Field Address
    //
    //        points to the previous definition the last definition is pointed
    //        to through the variable LAST.
    //
    //  FFA : + 4 :: Flag Field Address
    //
    //        byte 3 : always zero (hardware does that)
    //        byte 2 : always FF
    //        byte 1 : always FF
    //        byte 0 : bit 0 : immediate
    //                 bit 1 : hidden
    //
    //  NFA : + 8 :: Name Field Address
    //
    //        a variable lenght field, the low byte of the first pword
    //        holds the length, the remaining part holds the full name.
    //        When this does not completely fill out up to the next even
    //        flash address some alignment bytes are added.
    //
    //  CFA : + 8 + Length( Name) TCell aligned :: Code Field Address
    //
    //        This field always holds code, which can be the start of
    //        a code block for a primary word, or a jump to a handler
    //        routine for secondary words. Primaries will end with
    //        a goto Next jump usually.
    //
    //  PFA : + 12 + Length( Name) TCell aligned :: Parameter Field Address
    //
    //        For secondary words only, contains the data used by the handler
    //        compiled into the CFA. I.e. a list of pointers for a colon word
    //        (doCol handler), a value for a constant word (doCon) a pointer
    //        to RAM for a variable word (doVar) etc.
    //
    //        For deferred words there are two compiled fields, the first one
    //        pointing to the RAM location holding it's current value and the
    //        second one pointing to the previously defined deferred word.
    //
    //        The target variable lastdefer will point to the last deferred
    //        word.
    //
    //
    //  Some code snippets do not have a header, the system will 'magically'
    //  know where they are, and they will be compiled by higher level words
    //  when they are needed. The word constant will compile doCon, Variable
    //  will compile doVar and : (colon) will compile doCol etc. etc.
    //
    //  The current headerless words are :
    //
    //    doCol (or ENTER) - pointer list interpreter
    //    doCon            - constant interpreter
    //    doString         - string constant interpreter
    //    doVar            - variable interpreter
    //    doArray          - array interpreter
    //    doDefer          - defer interpreter
    //    doUser           - user variable interpreter
    //    doDoes>          - does> interpreter
    //
    // ////////////////////////////////////////////////////////////////////////
    //
  #endDoc


private

  uses
    Queues Utils Strings
  enduses

  //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  queue forth.input 512

  byteptr forth.index


  //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .


  {
    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    // Some meta magic
    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    also forth definitions


    2 constant THalfCell                            #// A Target HALFCELL is two  bytes (register size)
    4 constant TCell                                #// A Target CELL     is four bytes

    : TCells ( d -- d )

      TCell *
    ;

    : THalfCells ( d -- d )

      THalfCell *
    ;

    : debug ( n msg -- n )

      " : " + over >$ + message
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    variable DefinitionName                         #// Local  name for current definition
    variable TargetName                             #// Target name for current definition
    variable FlagsField                             #// Flags to be applied to current definition
    variable StackImage                             #// Target stack image for current definition
    variable LastLink         " 0" LastLink  !      #// Last target NFA value (symbolic)
    variable TVoffset           0  TvOffset  !      #// Current byte offset in target RAM "dictionary" space.
    variable TLastDefer       " 0" TLastDefer !     #// Target pointer to last deferred word (symbolic)


    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : TVAllot ( n -- n )

      #// Allort bn cells in target RAM and return it's byte offset

      TCells                                        // make n bytes from n Cells
      TVoffset @                                    // ( n noldSize ) in bytes
      swap over                                     // ( noldsize n noldsize )
      +                                             // ( noldsize nnewSize )
      TVoffset !                                    // ( noldsize )
      // " TVAllot_2" debug
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : TVCells ( -- n )

      #// Amount of target data cells currently allocated

      TVOffset @
      TCell /
    ;


    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : HeaderName ( -- str )

      #// Return the target header name from the definition name
      #// The definition name is the target name prefixed with t.

      DefinitionName @ " .h" +
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : registerHeader ( -- )

      #// Parse some tokens from the input and remember those.

      bl token TargetName !                           // Get and store the target name
      bl Token FlagsField !                           // Get and store the flags
      bl token StackImage !                           // Get and store stack image
      " t." TargetName @ + DefinitionName !           // Make a local name by prefixing t.
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : definition ( -- )

      #// Start a target definition by buildin a header into a string variable
      #// codeBuffer (as source code to be intepreted later on).

      registerHeader

      " " codeBuffer !
      "   code " HeaderName + "  " + StackImage @ +                                 cb+cr+
      "  "                                                                          cb+cr+
      "   .pword      "  LastLink   @ +                                             cb+cr+
      "   .pword      "  FlagsField @ +                                             cb+cr+
      "   .pascii     <" TargetName @ strLen >$ + " >,\"" + TargetName @ + " \"" +  cb+cr+
      "   .palign     2"                                                            cb+cr+
      "  "                                                                          cb+cr+

      "   .ifdef      __C30ELF"                                                     cb+cr+
      "   .type       <!\" " DefinitionName @ + " \" mangle>" + " , @function" +    cb+cr+
      "   .endif"                                                                   cb+cr+
      "  "                                                                          cb+cr+

      " <!\" " DefinitionName @ + " \" mangle>:" +                                  cb+cr+

      " <*\" " HeaderName + " \">" + LastLink !
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    todo' endDefinition seems to return twice? stuff after evaluate not executed ????'

    : endDefinition ( -- )

      #// End a target definition and compile it by interpreting the contents
      #// from the string variable codeBuffer

      "   endCode used"                                                             cb+cr+
      CodeBuffer @
      // dup message                                // debug code to dump code created into console
      evaluate
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : scanForCode ( str -- )

      #// Scan the input line by line until a line is found that contains
      #// str, just add the lines to the codeBuffer for later interpretation.

      begin
        readLine                                    // ( str line )
        over over                                   // ( str line str line )
        strPos 0 =
      while
        cb+cr+
      repeat
      skipToEOL
      "  " cb+
      drop
      drop
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.code ( -- )

      #// Start a target code word (primary) to be ended with t.endCode

      definition
        " t.endCode" scanForCode
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.: ( -- )

      #// Start a target colon word (secondary) to be ended with t.;

      definition
        "     bra       <*\" t.doCol\">                         ; // DOCOL :: Execute following pointer list" cb+cr+

        " t.;" scanForCode

        " "                                                                                                   cb+cr+
        "     .pword    <*\" t.semi\">                          ; // SEMI :: return to calling pointer list"  cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.constant ( value -- )

      #// Define a target constant

      definition
        "     bra       <*\" t.doCon\">                         ; // DOCON :: push following pword" cb+cr+
        "     .pword    " swap >$ +                                                                 cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.sym_constant ( value -- )

      #// Define a target symbolic constant

      definition
        "     bra       <*\" t.doCon\">                         ; // DOCON :: push following pword" cb+cr+
        "     .pword    " swap +                                                                    cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.string ( value -- )

      #// Define a target symbolic constant

      definition
        "     bra       <*\" t.doString\">                         ; // DOSTRING :: push following pword" cb+cr+
        "     .pascii   <" over strlen >$ + " >,\"" + swap + " \"" +                                      cb+cr+
        "     .palign   2"                                                                                cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.variable ( -- )

      #// Define a target variable

      1 TVAllot                                     // Make room in target RAM, return offset for it
      definition
        "     bra       <*\" t.doVar\">                         ; // DOVAR :: push following pword" cb+cr+
        "     .pword    <@\" VBank\"> + " swap >$ +                                                 cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.array ( n -- )

      #// Define a target array of n cells

      TVAllot                                       // Make room in target RAM, return offset for it
      definition
        "     bra       <*\" t.doArray\">                       ; // DOARRAY :: push following pword" cb+cr+
        "     .pword    <@\" VBank\"> + " swap >$ +                                                   cb+cr+
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    : t.defer ( -- )

      #// Define a target deferred word (code pointer).

      1 TVAllot                                     // Make room in target RAM, return offset for it
      definition
        "     bra       <*\" t.doDefer\">                       ; // DODEFER :: execute following pword"  cb+cr+
        "     .pword    <@\" VBank\"> + " swap >$ +                                                       cb+cr+
        "     .pword    "  TLastDefer @ + "                        ; // link to previous deferred word" + cb+cr+

        " <*\" " HeaderName + " \">" + TLastDefer !   // Make this definition the last deferred one
      endDefinition
    ;

    //  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    // End of meta stuff
  }


  // //////////////////////////////////////////////////////////////////////////
  // Set up some target definitions

20 12 + THalfCells  constant  RBankSize             #// Register  Bank  Size - in bytes (keep a few spares to allow for PS underflow)
    128 TCells      constant  PSSize                #// Parameter Stack Size - in bytes
    128 TCells      constant  RSSize                #// Return    Stack Size - in bytes
   4096 TCells      constant  VBankSize             #// Variable  Bank  Size - in bytes

        RBankSize   array     RBank                 #// Register bank storage area
        PSSize      array     PS                    #// Parameter Stack - TOS is at the lowest  address
        RSSize      array     RS                    #// Return    Stack - TOS is at the highest address
        VBankSize   array     VBank                 #// Variable bank storage area

                    2variable TVector               #// To set a target forth xt from native 4th - not implemented yet


  // //////////////////////////////////////////////////////////////////////////

  code forth.start ( -- )

    ; // Start the forth engine by jumping into COLD.
    ; //
    ; // Save registers
    ; //
    ; // This is a native 4th word which sets up the target forth
    ; // intepreter.

    breakpoint  forth.start
                                    ; // forth.start
    mov       w0, <@" RBank">       ; //  0 ::  0 : old w0
    mov       #<@" RBank"> + 2, w0
    mov       w1 , [w0++]           ; //  2 ::  1
    mov       w2 , [w0++]           ; //  4 ::  2
    mov       w3 , [w0++]           ; //  6 ::  3
    mov       w4 , [w0++]           ; //  8 ::  4
    mov       w5 , [w0++]           ; // 10 ::  5
    mov       w6 , [w0++]           ; // 12 ::  6
    mov       w7 , [w0++]           ; // 14 ::  7
    mov       w8 , [w0++]           ; // 16 ::  8
    mov       w9 , [w0++]           ; // 18 ::  9
    mov       w10, [w0++]           ; // 20 :: 10
    mov       w11, [w0++]           ; // 22 :: 11
    mov       w12, [w0++]           ; // 24 :: 12
    mov       w13, [w0++]           ; // 26 :: 13
    mov       w14, [w0++]           ; // 28 :: 14
    mov       w15, [w0++]           ; // 30 :: 15
    mov       TBLPAG, w2
    mov       w2 , [w0++]           ; // 32 :: 16
    mov       RCOUNT, w2
    mov       w2 , [w0++]           ; // 34 :: 17
    mov       SR, w2
    mov       w2 , [w0]             ; // 36 :: 18

    ; // Initialize forth interpreter

    mov       #<@" PS">, w9         ; // Set PSP to lowest  address in PS

    mov       #<@" RS">, w3         ; // Set RSP one beyond highest address in RS
    mov       #<@" RSSize">, w10
    add       w3, w10, w10

    ; // Jump into COLD, setting up initial W contents on the fly

    mov       #tblpage  ( <*" t.cold">), w2 ; // Set up t.cold as the entry point
    mov       w2, TBLPAG                    ; // Which needs be a forth secondary word
    mov       #tbloffset( <*" t.cold">), w8

    mov       w8    , w6            ; // Set up initial W to point to t.cold
    mov       TBLPAG, w7

    push      w6                    ; // JUMP IP (into t.cold)
    push      w7

    return
  endcode

  #document forth.start
    // Start the forth engine by jumping into COLD.
    // COLD will run some code and when finished it
    // will return control to the caller of START
    // trhough the word BYE.
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.firstdef ( -- )

  .palign     512                   ; // Align to a flash erase page boundary

  .global     <*" BEGIN_FORTH_FLASH">

<*" BEGIN_FORTH_FLASH">:

  endCode

  // //////////////////////////////////////////////////////////////////////////

  t.code bye 0xffff00 <( -- )>
                                    ; // bye
    ; // Restore registers

    mov       #<@" RBank"> + 36, w0
    mov       [w0--], w2            ; // 36 :: 18
    mov       w2, SR
    mov       [w0--], w2            ; // 34 :: 17
    mov       w2, RCOUNT
    mov       [w0--], w2            ; // 32 :: 16
    mov       w2, TBLPAG
    mov       [w0--], w15           ; // 20 :: 15
    mov       [w0--], w14           ; // 28 :: 14
    mov       [w0--], w13           ; // 26 :: 13
    mov       [w0--], w12           ; // 24 :: 12
    mov       [w0--], w11           ; // 22 :: 11
    mov       [w0--], w10           ; // 20 :: 10
    mov       [w0--], w9            ; // 18 ::  9
    mov       [w0--], w8            ; // 16 ::  8
    mov       [w0--], w7            ; // 14 ::  7
    mov       [w0--], w6            ; // 12 ::  6
    mov       [w0--], w5            ; // 10 ::  5
    mov       [w0--], w4            ; //  8 ::  4
    mov       [w0--], w3            ; //  6 ::  3
    mov       [w0--], w2            ; //  4 ::  2
    mov       [w0--], w1            ; //  2 ::  1
    mov       [w0]  , w0            ; //  0 ::  0

    ; // Return to caller of forth.start

    breakpoint  t.bye

    return
  t.endCode

  #document t.bye.h
    // CODE BYE ( -- )
    //
    // Returns to the caller of START
  #endDoc


  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  //
  // Headerless words for inner interpreter

  code t.doCol ( -- ) // [ -- d ] (or ENTER)
                                    ; // doCol
    mov       w8, [--w10]           ; // IP PUSH RSP
    mov       TBLPAG, w2
    mov       w2, [--w10]

    inc2      w6, w8                ; // PFA TO IP
    addc      w7, #0, w2
    mov       w2, TBLPAG

<*" t.next">:

    tblrdl.w  [w8  ], w6            ; // [IP] TO W   CELL +TO IP ; // [IP++]->W,
    tblrdh.w  [w8++], w7

    push      w6                    ; // W JUMP
    push      w7
    return
  endCode used

  #document t.doCol
    // DOCOL :: A code fragment, not a forth primary
    //   IP PUSH RSP
    //   PFA TO IP
    //   NEXT
    //
    // NEXT :: A code fragment, not a forth primary
    //   [IP] TO W
    //   CELL +TO IP
    //   W JUMP
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doCon ( -- d )
                                    ; // doCon
    mov       w0, [++w9]            ; // [PFA] PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w6                ; // PFA
    addc      w7, #0, w7
    mov       w7, TBLPAG

    tblrdl.w  [w6], w0              ; // [PFA]
    tblrdh.w  [w6], w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doCon
    // A code fragment, not a forth primary
    //   [PFA] PUSH PSP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doString ( -- d )
                                    ; // doCon
    mov       w0, [++w9]            ; // PFA PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w0
    addc      w7, #0, w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doCon
    // A code fragment, not a forth primary
    //   PFA PUSH PSP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doVar ( -- d )
                                    ; // doVar
    mov       w0, [++w9]            ; // [PFA] PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w6                ; // PFA    // PFA holds pointer to RAM
    addc      w7, #0, w7
    mov       w7, TBLPAG

    tblrdl.w  [w6], w0              ; // [PFA]
    tblrdh.w  [w6], w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doVar
    // A code fragment, not a forth primary
    //   [PFA] PUSH PSP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doArray ( -- d )
                                    ; // doArray
    mov       w0, [++w9]            ; // [PFA] PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w6                ; // PFA
    addc      w7, #0, w7
    mov       w7, TBLPAG

    tblrdl.w  [w6], w0              ; // [PFA]
    tblrdh.w  [w6], w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doArray
    // A code fragment, not a forth primary
    //   [PFA] PUSH PSP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doDefer ( -- )
                                    ; // doDefer
    mov       w0, [++w9]            ; // [PFA] PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w6                ; // PFA
    addc      w7, #0, w7
    mov       w7, TBLPAG

    tblrdl.w  [w6], w0              ; // [PFA]
    tblrdh.w  [w6], w1

    mov       [w0++], w2            ; // @
    mov       [w0  ], w1
    mov       w2, w0

    goto      <*" t.execute">       ; // EXECUTE
  endCode used

  #document t.doDefer
    // A code fragment, not a forth primary
    //   [PFA] PUSH PSP
    //   @
    //   EXECUTE
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doUser ( -- d )
                                    ; // doUser
    breakpoint UNTESTED             // !!! This code was not tested yet !!!

    mov       w0, [++w9]            ; // [PFA] PUSH PSP
    mov       w1, [++w9]

    inc2      w6, w6                ; // PFA
    addc      w7, #0, w7
    mov       w7, TBLPAG

    tblrdl.w  [w6], w0              ; // [PFA]
    tblrdh.w  [w6], w1

    mov       <@" VBank"> + 4, w2   ; // UP
    mov       <@" VBank"> + 6, w3

    add       w1, w3, w1            ; // +
    add       w0, w2, w0
    addc      w1, #0, w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doUser
    // A code fragment, not a forth primary
    //   [PFA] UP + PUSH PSP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.doDoes> ( -- d )
                                    ; // doDoes>
    mov       w8, [--w10]           ; // IP PUSH RSP
    mov       TBLPAG, w2
    mov       w2, [--w10]

    pop       w3                    ; // POP [hardware call stack] TO IP
    pop       w2
    mov       w3, TBLPAG

    tblrdl.w  [w2], w8
    tblrdh.w  [w2], w4
    mov       w4, TBLPAG

    mov       w1, [++w9]            ; // PFA -> TOS
    mov       w0, [++w9]

    inc2      w2, w0
    addc      w3, #0, w1

    goto      <*" t.next">          ; // NEXT
  endCode used

  #document t.doDoes>
    // A code fragment, not a forth primary
    //
    // Enters a high-level thread with PFA on TOS.
    //
    //   IP PUSH RSP                        // Push old IP on return stack
    //   POP [hardware call stack] TO IP    // address of new thread to IP
    //   PFA PUSH PSP                       // high level code expects PFA on TOS
    //   NEXT                               // Go into high level code after DOES>
    //
    // Or .. from http://www.jimbrooks.org/web/forth/forthInternals.php
    //
    //   Where DODOE is doDoes> and <BUILDS is CREATE
    //
    //   DODOE first pushes IP to the R-stack, as done with colon definitions.
    //   DODOE then loads IP with CFA held in the PFA, and pushes the address -after- the PFA,
    //   i.e., the cell immediately after the PFA. This address points to the start of the
    //   data compiled by any compiling words specified after <BUILDS.
    //   DODOE then executes NEXT to execute the CFA held in the PFA.
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code forth.leaveto.4th ( -- )

    breakpoint UNTESTED             // !!! This code was not tested yet !!!

    mov       w0, [++w9]            ; // Push low word onto 4th stack
    mov       w1, w0                ; // And high word into TOS cache

    goto      <*" t.next">          ; // NEXT
  endCode used
  #document forth.leaveto.4th
    // forth helper code to exit back to forth from a 4th call
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code 4th.call.forth ( dxt -- )

    breakpoint UNTESTED             // !!! This code was not tested yet !!!

    ; // Jump into xt, setting up initial W contents on the fly

    mov       w0, w1                ; // Set up xt as the entry point, Which needs be a forth secondary word
    mov       w1, TBLPAG
    mov       [w9--], w8

    mov       w8, w6                ; // Set up initial W to point to xt
    mov       w1, w7

    mov       #tblpage  ( <*" forth.leaveto.4th">), w1
    mov       #tbloffset( <*" forth.leaveto.4th">), w0

    mov       w0, [--w10]           ; // Push 32 bit IP forth.leaveto.4th on return stack / 16 bit loop stack
    mov       w1, [--w10]

    mov       [w9--], w1            ; // Set up TOS cache for forth code, pop w1,w0
    mov       [w9--], w0

    push      w6                    ; // JUMP IP (into xt)
    push      w7

    return
  endCode used
  #document 4th.call.forth
    // Call 32 bit forth from 16 bit 4th, which will return through its Next code
    // so we need push some IP here
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  macro call.forth ( aName ) ( ? -- ? )

    PushLit   tbloffset( <*" t.$aName$">)
    PushLit   tblpage  ( <*" t.$aName$">)
    mov       #0x00ff, w1
    and       w0, w1, w0
    call      <*" 4th.call.forth">

  endMacro


  // //////////////////////////////////////////////////////////////////////////
  // //////////////////////////////////////////////////////////////////////////
  //
  // Inner interpreter words with headers, and some helper routines

  // //////////////////////////////////////////////////////////////////////////

  code forth.call.4th ( dxt -- )

    breakpoint UNTESTED             // !!! This code was not tested yet !!!

    mov       w0, w6                ; // xt to w7,w6
    mov       w1, w7

    mov       [w9--], w0            ; // pop TOS to w0

    mov       w8, [--w10]           ; // Push 32 bit IP on return stack / 16 bit loop stack
    mov       TBLPAG, w8
    mov       w8, [--w10]

    push      w6                    ; // Call into xt, which returns trough a return to our caller
    push      w7

    return
  endCode used
  #document forth.call.4th
    // Call some 16 bit 4th from 32 bit forth and then return to caller
    //
    // Helper for forth>4th
    //
    // A code fragment, not a forth primary
    // This will do an actual call to a 4th xt with whatever is
    // on the forth stacks, so that needs be preserved by the code
    // at xt. When the xt code returns it returns to this code's
     // caller (presumably forth.run.4th)
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code forth>4th 0xffff00 <( dxt -- )>

    breakpoint UNTESTED             // !!! This code was not tested yet !!!

    call      <*" forth.call.4th"> ;

    mov       [w10++], w8           ; // Pop 32 bit IP from return stack / 16 bit loop stack
    mov       w8, TBLPAG
    mov       [w10++], w8

    mov       w0, w1                ; // Move high word to w1
    mov       [w9--], w0            ; // Pop low word into 32 byte cache register's w0

    goto      <*" t.next">          ; // NEXT
  t.endCode
  #document t.forth>4th.h
    // Call into 16 bit 4th code located at xt, passing the stacks
    // to be usable by the 16 bit 4th code
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code execute 0xffff00 <( xt -- )>
                                    ; // execute
    mov       w0, w6                ; // POP PSP TO W
    mov       w1, w7
    mov       [w9--], w1
    mov       [w9--], w0

    push      w6                    ; // W JUMP
    push      w7
    return
  t.endCode

  #document t.execute.h
    // CODE EXECUTE
    //   POP PSP TO W
    //   W JUMP
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: perform 0xffff00 <( n -- )>
                                    ; // perform
    .pword    <*" t.@">             ; // @ execute
    .pword    <*" t.execute">
  t.;

  #document t.perform.h
    // : PERFORM @ EXECUTE ;
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code semi 0xffff00 <( -- ) // [ d -- ] or EXIT>
                                    ; // semi
    mov       [w10++], w2           ; // POP RSP TO IP
    mov       w2, TBLPAG
    mov       [w10++], w8

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.semi.h
    // CODE SEMI
    //   POP RSP TO IP
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code doLit 0xffff00 <( -- d )>
                                    ; // doLit
    mov       w0, [++w9]            ; // Push next inline data word onto data stack
    mov       w1, [++w9]
    tblrdl.w  [w8  ], w0
    tblrdh.w  [w8++], w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.doLit.h
    // CODE DOLIT
    //   push inline value
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code breakpoint 0xffff00 <( -- )>
                                    ; // breakpoint
    breakpoint  t.breakpoint

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.breakpoint.h
    // CODE BREAKPOINT
    //   stop execution here in the debugger
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code cfa2pfa 0xffff00 <( d -- d )>
                                    ; // cfa2pfa
    add       w0, #2, w0
    addc      w1, #0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.cfa2pfa.h
    // : cfa2pfa ( d -- d )
    //   calculate PFA from CFA
    // ;
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code 0= 0xffff00 <( d -- df )>
                                    ; // 0=
    ior       w0, w1, w0            ; // Or low and high word
    sub       #1, w0                ; // Subtract one, borrow set if 0= holds
    subb      w0, w0, w0            ; // when borrow set $ffffffff, 0 otherwise
    mov       w0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.0=.h
    // code 0= ( d -- df )
    //   return $ffffffff when d = 0, and 0 otherwise
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code branch 0xffff00 <( -- )>
                                    ; // branch
    tblrdl.w  [w8], w2              ; // [IP] -> IP
    tblrdh.w  [w8], w3
    mov       w2, w8
    mov       w3, TBLPAG

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.branch.h
    // code branch ( -- )
    //   branch to the following inline absolute address
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code fbranch 0xffff00 <( df -- )>
                                    ; // fbranch
    ior       w0, w1, w0            ; // Or low and high word, branch if zero

    bra       z, 1f                 ; // B/ calculate new IP (branch)
    bra       2f                    ; // B/ Skip over current IP

1:

    tblrdl.w  [w8], w2              ; // [IP] -> IP // Branch
    tblrdh.w  [w8], w3
    mov       w2, w8
    mov       w3, TBLPAG
    bra       3f

2:

    mov       TBLPAG, w2            ; // IP++       // No branch
    add       w8, #2, w8
    addc      w2, #0, w2
    mov       w2, TBLPAG

3:

    mov       [w9--], w1            ; // Drop df
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.fbranch.h
    // code fbranch
    //   branch to the following inline absolute address
    //   when TOS = 0, skip that address otherwise (and
    //   do not branch then).
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code drop 0xffff00 <( d -- )>
                                    ; // drop
    mov       [w9--], w1
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.drop.h
    // CODE DROP
    //   drop TOS
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code dup 0xffff00 <( d -- d d )>
                                    ; // dup
    mov       w0, [++w9]
    mov       w1, [++w9]

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.dup.h
    // CODE DUP
    //   dup TOS so NOS and TOS are equal
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code >r 0xffff00 <( d -- ) // [ -- d ]>

    mov       w0, [--w10]           ; // push onto return stack
    mov       w1, [--w10]

    mov       [w9--], w1            ; // Pop from data stack
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.>r.h
    // CODE >R
    //   move TOS to TOR
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code r> 0xffff00 <( -- d ) // [ d -- ]>

    mov       w1, [++w9]            ; // Push to data stack
    mov       w0, [++w9]

    mov       [w10++], w0           ; // Pop from return stack
    mov       [w10++], w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.r>.h
    // CODE >R
    //   move TOR to TOS
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code c@ 0xffff00 <( dAddress -- dData )>
                                    ; // c@
    sl        w0, w2                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left
    bra       z, 1f

    mov       w1, DSRPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov.b     [w0], w0              ; // Get the byte
    clr       w1                    ; // Clear high word
    and       #0xff, w0             ; // mask out high byte of low word

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.c@.h
    // CODE C@ ( d -- d ) // fetch
    //  get dData from dAddress
    //  NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code c! 0xffff00 <( dData dAddress -- )>
                                    ; // c!
    mov       [w9--], w3            ; // dData -> w3,w2 - high word ignored      (pops)
    mov       [w9--], w2            ; //                  high byte ignored too

    sl        w0, w4                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left
    bra       z, 1f

    mov       w1, DSWPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov.b     w2, [w0]              ; // Write the byte

    mov       [w9--], w1            ; // And drop an item
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.c!.h
    // CODE C! ( d d -- ) // store
    //   store dData at dAddress
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code w@ 0xffff00 <( dAddress -- dData )>
                                    ; // w@
    sl        w0, w2                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left
    bra       z, 1f

    mov       w1, DSRPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov       [w0], w0              ; // Get the word
    clr       w1                    ; // Clear high word

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.w@.h
    // CODE W@ ( d -- d ) // fetch
    //  get dData from dAddress
    //  NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code w! 0xffff00 <( dData dAddress -- )>
                                    ; // w!
    mov       [w9--], w3            ; // dData -> w3,w2 - high word ignored (pop)
    mov       [w9--], w2

    sl        w0, w4                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left
    bra       z, 1f

    mov       w1, DSWPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov       w2, [w0]              ; // Hmm this supports only 64 k of RAM

    mov       [w9--], w1            ; // And drop an item
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.w!.h
    // CODE W! ( d d -- ) // store
    //   store dData at dAddress
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code @ 0xffff00 <( dAddress -- dData )>
                                    ; // @
    sl        w0, w2                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left
    bra       z, 1f

    mov       w1, DSRPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov       [w0++], w2            ; // Get the long word .. it must not cross an EDS page boundary ...
    mov       [w0  ], w1
    mov       w2, w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.@.h
    // CODE @ ( d -- d ) // fetch
    //  get dData from dAddress
    //  NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code ! 0xffff00 <( dData dAddress -- )>

    ; // NOTE : write can not cross an EDS page boundary
                                    ; // !
    mov       [w9--], w3            ; // dData -> w3,w2 (pop)
    mov       [w9--], w2

    sl        w0, w4                ; // Shift left low address into w2 over one bit, bit 15 into carry
    rlc       w1, w1                ; // Shift carry into w1 and w1 one bit left

    bra       z, 1f
    mov       w1, DSWPAG            ; // Set up read page
    bset      w0, #15               ; // Enable EDS

1:

    mov       w2, [w0++]            ; // Hmm this supports only 64 k of RAM
    mov       w3, [w0  ]            ; // ... anyway ... move dData to dAddress

    mov       [w9--], w1            ; // And drop an item
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.!.h
    // CODE ! ( d d -- ) // store
    //   store dData at dAddress
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code p@ 0xffff00 <( d -- d )>
                                    ; // p@
    mov       TBLPAG, w3            ; // Save TBLPAG
    mov       w1, TBLPAG            ; // Set up flash address into TBLPAG, w2
    mov       w0, w2
    tblrdl.w  [w2], w0              ; // Read two words from flash
    tblrdh.w  [w2], w1
    mov       w3, TBLPAG            ; // Restore TBLPAG

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.p@.h
    // code p@ ( d -- d )
    //   fetch a value from code space (flash / program space)
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code + 0xffff00 <( d1 d2 -- d )>
                                    ; // +
    add       w1, [w9--], w1        ; // Add d1 and d2, dropping d2, replacing d1 by sum
    add       w0, [w9--], w0
    addc      w1, #0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.+.h
    // CODE +
    //   add d1 and d2
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code negate 0xffff00 <( d -- d )>
                                    ; // negate
    neg     w1, w1                  ; // Two's complement
    neg     w0, w0
    subb    #0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.negate.h
    // CODE negate
    //   negate TOS
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: - 0xffff00 <( d1 d2 -- d )>
                                    ; // -
    .pword    <*" t.negate">        ; // negate
    .pword    <*" t.+">             ; // +
  t.;

  #document t.-.h
    // : - ( d1 d2 -- d ) negate + ;
    //   subtract d2 from d1
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code and 0xffff00 <( d1 d2 -- d )>
                                    ; // and
    and       w1, [w9--], w1        ; // And d1 and d2, popping d2, replacing d1 with and
    and       w0, [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.and.h
    // CODE and
    //   and d1 and d2
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code or 0xffff00 <( d1 d2 -- d )>
                                    ; // or
    ior       w1, [w9--], w1        ; // Or d1 and d2, popping d2, replacing d1 with or
    ior       w0, [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.or.h
    // CODE or
    //   or d1 and d2
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code xor 0xffff00 <( d1 d2 -- d )>
                                    ; // xor
    xor       w1, [w9--], w1        ; // Xor d1 and d2, popping d2, replacing d1 with xor
    xor       w0, [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.xor.h
    // CODE xor
    //   or d1 and d2
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code not 0xffff00 <( d -- d )>
                                    ; // not
    com       w0, w0                ; // One's complement
    com       w1, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.xor.h
    // CODE xor
    //   or d1 and d2
    //   NEXT
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code cells 0xffff00 <( d -- d )>
                                    ; // cells
    sl      w0, w0                  ; // Shift left low word,
                                    ; // shift a zero to bit 0
                                    ; // shift bit 15 into carry
    rlc     w1, w1                  ; // rotate left high word
                                    ; // carry into bit 0, bit 15 into carry

    sl      w0, w0                  ; // And again
    rlc     w1, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.cells.h
    // : CELLS ( d -- d ) 2* 2* ;
    //   multiply d by CELL
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: = 0xffff00 <( d1 d2 -- df )>
                                    ; // =
    .pword    <*" t.-">             ; // - 0=
    .pword    <*" t.0=">
  t.;

  #document t.=.h
    // : = - 0= ; // return true when d1 = d2, false otherwise
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: =marker 0xffff00 <( d -- df )>
                                    ; // =marker
    .pword    <*" t.marker.mask">   ; // marker.mask and marker.mask =
    .pword    <*" t.and">
    .pword    <*" t.marker.mask">
    .pword    <*" t.=">
  t.;

  #document t.=marker.h
    // : =marker marker.mask and marker.mask = ; // return true when d is a header marker, false otherwise
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code 2+ 0xffff00 <( d -- d )>
                                    ; // 2+
    add       w0, #2, w0
    addc      w1, #0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.2+.h
    // code 2_ ( d - d )
    //   add 2 to TOS
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code 2- 0xffff00 <( d -- d )>
                                    ; // 2-
    sub       w0, #2, w0
    subb      w1, #0, w1

    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.2-.h
    // code 2_ ( d - d )
    //   subtract 2 from TOS
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: cfa2ffa 0xffff00 <( d -- d )>
                                    ; // cfa2ffa
<*" cfa2ffa_loop">:                 ; // begin

    .pword    <*" t.2-">            ; //   2-
    .pword    <*" t.dup">           ; //   dup
    .pword    <*" t.p@">            ; //   p@
    .pword    <*" t.=marker">       ; //   =marker
    .pword    <*" t.fbranch">       ; // until
    .pword    <*" cfa2ffa_loop">    ; //
  t.;

  #document t.cfa2ffa.h
    // : cfa2ffa ( d -- d )            // Determine Flag Field Address (FFA) from CFA
    //   begin
    //     2- dup p@ =marker           // Serarch backwards from CFA till marker found
    //   until
    // ;
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: cfa2lfa 0xffff00 <( d -- d )>
                                    ; // cfa2lfa
    .pword    <*" t.cfa2ffa">       ; // cfa2ffa 2-
    .pword    <*" t.2-">
  t.;

  #document t.cfa2lfa.h
    // : cfa2lfa ( d -- d ) cfa2ffa 2- ; // Get Link Field Address from CFA
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: cfa2nfa 0xffff00 <( d -- d )>
                                    ; // cfa2nfa
    .pword    <*" t.cfa2ffa">       ; // cfa2ffa 2+
    .pword    <*" t.2+">
  t.;

  #document t.cfa2nfa.h
    // : cfa2nfa ( d -- d ) cfa2ffa 2+ ; // Get Name Field Address from CFA
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: vallot 0xffff00 <( d -- )>
                                    ; // vallot
    .pword    <*" t.cells">         ; // CELLS -> bytes
    .pword    <*" t.vhere">         ; // vhere
    .pword    <*" t.@">             ; // @
    .pword    <*" t.+">             ; // +
    .pword    <*" t.vhere">         ; // vhere
    .pword    <*" t.!">             ; // !
  t.;

  #document t.vallot.h
    // : VALLOT ( d -- ) CELLS VHERE @ + VHERE ! ; // allot d cells to variable space
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: (is) 0xffff00 <( dtoken ddefer -- )>
                                    ; // (is)
    .pword    <*" t.cfa2pfa">       ; // cfa2pfa p@ !
    .pword    <*" t.p@">
    .pword    <*" t.!">
  t.;

  #document t.(is).h
    // : (is) ( dtoken ddefer -- ) cfa2pfa p@ ! ;
    // Resolve a deferred word with execution token ddefer to dtokenq
    // i.e. set the deferred word to execute dtoken
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.code noop 0xffff00 <( -- )>
                                    ; // noop
    goto      <*" t.next">          ; // NEXT
  t.endCode

  #document t.noop.h
    // code noop ( -- )
    //   wastes a few cycles
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  t.: definitions 0xffff00 <( -- )>
                                    ; // definitions
    .pword    <*" t.context">       ; // context @ current!
    .pword    <*" t.@">
    .pword    <*" t.current">
    .pword    <*" t.!">
  t.;

  #document t.definitions.h
    // : definitions ( -- ) context @ current ! ;
    // make compiler vocabulary (current) to the currently topmost
    // search vocabulary (context)
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

      t.variable  vhere       0xffff00 <( -- d )>       // points to first free RAM   location
      t.variable  chere       0xffff00 <( -- d )>       // points to first free FLASH location
      t.variable  marker      0xffff00 <( -- d )>       // points to point beyond which forget is not possible
      t.variable  last        0xffff00 <( -- d )>       // Points to the latest defined Name Field
      t.variable  lastdefer   0xffff00 <( -- d )>       // Points to the latest deferred word's Name Field
      t.variable  memfree     0xffff00 <( -- d )>       // Bytes free in forth FLASH area
      t.variable  memsize     0xffff00 <( -- d )>       // Total size of forth FLASH area
      t.variable  current     0xffff00 <( -- d )>       // Vocabulary where new definitions will go
      t.variable  context     0xffff00 <( -- d )>       // Vocabulary used to look up words
      t.variable  up          0xffff00 <( -- d )>       // User Pointer
      t.variable  state       0xffff00 <( -- d )>       // Forth State, 0 = interpreting, 1 = compiling
      t.variable  separator   0xffff00 <( -- d )>       // Forth token scanner separator
      t.variable  parsestate  0xffff00 <( -- d )>       // Forth token scanner parse state
   64 t.array     parsebuf    0xffff00 <( -- d )>       // Forth token scanner parse buffer
      t.defer     (main)      0xffff00 <( -- d )>       // ?? hmm .. are we going to keep that ??
      t.defer     (test)      0xffff00 <( -- d )>       // ?? hmm .. are we going to keep that ??


  // //////////////////////////////////////////////////////////////////////////

          0 t.constant  false           0xffff00 <( -- d )>   // False
  $ffffffff t.constant  true            0xffff00 <( -- d )>   // True
          4 t.constant  cell            0xffff00 <( -- d )>   // Target CELL size
    TVCells t.constant  initial.vallot  0xffff00 <( -- d )>   // Define this after all t.variable and t.defer definitions (in cells)
  $00ffff00 t.constant  marker.mask     0xffff00 <( -- d )>   // Marker mask used to trace back FFA from CFA
        $20 t.constant  chrSpace        0xffff00 <( -- d )>   // Space character
        $0d t.constant  chrCR           0xffff00 <( -- d )>   // CR character
        $0a t.constant  chrLF           0xffff00 <( -- d )>   // LF character
" a string" t.string    astring         0xffff00 <( -- d )>   // String constant (in flash)

TLastDefer @ t.sym_constant last.defer  0xffff00 <( -- d )>   // Define this after all t.variable and t.defer definitions (in cells)


  // //////////////////////////////////////////////////////////////////////////

  // //////////////////////////////////////////////////////////////////////////

  t.code forth.accept 0xffff00 <( dc -- )>

    mov       [w9--], w1            ; // Drop df
    mov       [w9--], w0

    goto      <*" t.next">          ; // NEXT
  t.endCode


  // //////////////////////////////////////////////////////////////////////////

  // : test_does_var ( -- ) create 1 vallot does> p@ ;

  t.: create 0xffff00 <( -- )>

    ; // more ...

  t.;

  t.: does> 0xffff00 <( -- )>

    .pword    <*" t.r\>">           ; // R>

    ; // more ...

  t.;

  t.: test_does_var 0xffff00 <( -- )>

    .pword    <*" t.create">        ; // CREATE
    .pword    <*" t.doLit">         ; //   1 VALLOT
    .pword    1
    .pword    <*" t.vallot">
    .pword    <*" t.does\>">        ; // DOES>

<*" do_test_does_var">:

    .pword    <*" t.p@">            ; //   p@
  t.;

  // //////////////////////////////////////////////////////////////////////////

  // test_does_var aap

  t.code aap 0xffff00 <( -- n )>

      call      <* " t.doDoes>">            ; // Pushes PC + 4
      .pword    <*" do_test_does_var">      ; // which points here
      .pword    0xdeface                    ; // Address of this location in TOS after dodoes>
  t.endCode

  // //////////////////////////////////////////////////////////////////////////

  t.: cold 0xffff00 <( -- )>
                                    ; // cold
    .pword    <*" t.doLit">         ; // VBANK VHERE !
    .pword    <*" VBank">
    .pword    <*" t.vhere">
    .pword    <*" t.!">

    ; // VALLOT all defined target variables
    ; // NOTE :: use this after VHERE was setup

    .pword    <*" t.initial.vallot">  ; // INITIAL.VALLOT VALLOT
    .pword    <*" t.vallot">

    ; // Set CHERE to point to first free FLASH location

    .pword    <*" t.doLit">         ; // CURRENT_FORTH_FLASH DUP CHERE ! MARKER !
    .pword    <*" CURRENT_FORTH_FLASH">
    .pword    <*" t.dup">
    .pword    <*" t.chere">
    .pword    <*" t.!">
    .pword    <*" t.marker">
    .pword    <*" t.!">

    .pword    <*" t.doLit">         ; // FORTH_FLASH_FREE MEMFREE !
    .pword    <*" FORTH_FLASH_FREE">
    .pword    <*" t.memfree">
    .pword    <*" t.!">

    .pword    <*" t.doLit">         ; // FORTH_FLASH_SIZE MEMSIZE !
    .pword    <*" FORTH_FLASH_SIZE">
    .pword    <*" t.memsize">
    .pword    <*" t.!">


    ; // Set LAST to point to last header definition (this one)

    .pword    <*" t.doLit">         ; // ' COLD LAST !
    .pword    <*" t.cold.h">
    .pword    <*" t.last">
    .pword    <*" t.!">

    ; // Set LASTDEFER to point to the last deferred word

    .pword    <*" t.last.defer">    ; // LastDefer lastdefer ! last.defer
    .pword    <*" t.lastdefer">
    .pword    <*" t.!">

    .pword    <*" t.last">          ; // LAST @ CONTEXT !
    .pword    <*" t.@">
    .pword    <*" t.context">
    .pword    <*" t.!">

    .pword    <*" t.definitions">   ; // DEFINITIONS ( context @ current ! )

    .pword    <*" t.doLit">         ; // 0 UP !
    .pword    0
    .pword    <*" t.up">
    .pword    <*" t.!">

    .pword    <*" t.doLit">         ; // ' (MAIN) IS NOOP
    .pword    <*" t.noop">
    .pword    <*" t.doLit">
    .pword    <*" t.(main)">
    .pword    <*" t.(is)">

    ; // test code

    .pword    <*" t.aap">
    .pword    <*" t.drop">

    ; // rnd test code

    .pword    <*" t.(main)">        ; // (MAIN)

    .pword    <*" t.bye">           ; // BYE :: Return to caller of forth.start
  t.;

  #document t.cold.h
    // : COLD ( -- )
    //
    //   VBANK VHERE !
    //   0     UP    !
    //   BYE
    // ;
  #endDoc


  // //////////////////////////////////////////////////////////////////////////

  code t.lastdef ( -- )

  .global     <*" CURRENT_FORTH_FLASH">
  .global     <*" END_FORTH_FLASH">
  .global     <*" FORTH_FLASH_FREE">
  .global     <*" FORTH_FLASH_SIZE">

  .palign     2

<*" CURRENT_FORTH_FLASH">:
                                    ; // lastdef
  .pascii     "This is the end."

  .palign     512                   ; // Align to a flash erase page boundary


<*" END_FORTH_FLASH">:

  .equiv      <*" FORTH_FLASH_FREE">, <*" END_FORTH_FLASH"> - <*" CURRENT_FORTH_FLASH">
  .equiv      <*" FORTH_FLASH_SIZE">, <*" END_FORTH_FLASH"> - <*" BEGIN_FORTH_FLASH">

  endCode


  // //////////////////////////////////////////////////////////////////////////

  : forth.driver ( -- )

    #// Process queued data

    begin
      forth.input.hasData
    while
      forth.input.read 0                        // Read character, extend to 32 bits
      call.forth forth.accept                   // and handle it in forth with forth.accept
    repeat
  ;

  // //////////////////////////////////////////////////////////////////////////

  : forth.pre_driver ( -- )

    #// On queue overflows try process some data

    forth.input.hasRoom 0=
    if
      forth.driver
    endif
  ;


  // //////////////////////////////////////////////////////////////////////////

  : forth.$ram.accept ( $ram -- )

    #// Accept data from a RAM location

    >r forth.indexPtr!
    r> 0 ?do
      forth.pre_driver
      forth.indexc@++
      forth.input.write
    loop
    forth.driver
  ;

  // //////////////////////////////////////////////////////////////////////////


  : forth.rom.accept ( prom  -- )

    #// Accept data from a ROM location

    RomPtr!
    Romc@++
    0 ?do
      forth.pre_driver
      Romc@++
      forth.input.write
    loop
    forth.driver
  ;

  // //////////////////////////////////////////////////////////////////////////

  : forth.init ( -- )

    forth.input.clear
    forth.start
  ;
  resources
    t.firstdef
    t.lastdef
  endResources

  // //////////////////////////////////////////////////////////////////////////

public

  export  forth.init        (       --  )
  export  forth.$ram.accept ( $ram  --  )
  export  forth.rom.accept  ( prom  --  )

endModule

