
xVersion" 2012.10.10.1"

module state_machines

  #document state_machines

    // abstract state machines

    // A state machine has
    //
    //   State       : a state variable and a set of constant definitions to represent the state values
    //   Events      : "things" to throw at the machine that makes it perform a transition based on the
    //                 current state and the event "type"
    //   Transitions : <old_state, event, new_state, action>
    //
    // Notation :
    //
    //   StateMachine StateMachineName
    //
    //     transition state_old event state_new action
    //
    //
    //   EndStateMachine
    //
    //
    // Actions can be deferred words
    // Transitions can be implemented

  #enddoc


public

{

  // Meta definitions

  variable sm.machineName             // Name of current state machine being defined
  variable sm.states                  // Head of the list of state names
  variable sm.lastState               // Ordinal value for last state definition
  variable sm.events                  // Head of the list of event names
  variable sm.lastEvent               // Ordinal value for last event definition
  variable sm.transitions             // Head of the list of transitions


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : .spy ( n -- )

    " id     : " . . cr
    " stack  : " . .s
    " stream : " . bl token dup bl untoken . cr
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.clearStates ( -- )

    0 sm.states !                     // Clear Head of the list of state names
    0 sm.lastState !                  // Clear Ordinal value for last state definition
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.clearEvents ( -- )

    0 sm.events !                     // Clear Head of the list of event names
    0 sm.lastEvent !                  // Clear Ordinal value for last event definition
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.clearTransitions ( -- )

    0 sm.transitions !                // Clear Head of the list of transition names
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.prefix ( nStr nPrefix -- nPrefixedStr )

    swap +
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.prefix._ ( nStr -- nPrefixedStr )

    " -" sm.prefix
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.expand ( nStr nPrefix -- nPrefixedStr )

    sm.machineName @ sm.prefix sm.prefix
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.getToken ( -- nStr )

    bl token
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.pushToken ( nStr -- )

    bl unToken
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.sExpand ( nStr -- nPrefixedStr ) " _s_" sm.expand ;
  : sm.eExpand ( nStr -- nPrefixedStr ) " _e_" sm.expand ;
  : sm.tExpand ( nStr -- nPrefixedStr ) " _t_" sm.expand ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.findInList ( nStr nListPtr -- f )

    @
    false >r                          // ( nStr nList ) [ false ]
    begin                             // ( nStr nList ) [ false ]
      ?dup
      if                              // ( nStr nList^ ) [ false ]
        dup pfa@@ swap
        1 + pfa@@ 2 pick
        =
        if
          drop r> not >r              // Make result true
        endif
        r> dup >r                     // Stop loop when result true
      else
        true                          // Pointer was 0, not found, stop loop
      endif
    until
    drop                              // Drop nStr
    r>                                // Get result
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.findState?      ( nStr -- f ) sm.sExpand sm.states      sm.findInList ;
  : sm.findEvent?      ( nStr -- f ) sm.eExpand sm.events      sm.findInList ;
  : sm.findTransition? ( nStr -- f ) sm.tExpand sm.transitions sm.findInList ;


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.enclose ( n nvar -- )

    dup >r                            // dup nvar and keep taht copy on the return stack
    @ ,                               // Get value from nvar and compile into dictionary
    r> !                              // Get copy of nvar back and store n there
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.hereEnclose ( nvar -- )

    here swap sm.enclose              // Link here into a list point to by nvar
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.newSmConst ( nVar -- )

    dup >r                            // dup nvar to return stack
    @ dup constant                    // get current value and make a constant with that value
    1 + r> !                          // Increment value in nvar
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.newState ( -- ) sm.lastState sm.newSmConst ;
  : sm.newEvent ( -- ) sm.lastEvent sm.newSmConst ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.addState ( nStr -- )

    sm.sExpand                        // Build <machineName>_s_<nStr>
    dup dup sm.prefix._               // ( tok tok _tok )
    swap sm.pushToken                 // ( tok _tok )
    sm.newState                       // Define a new state constant named tok
    sm.pushToken                      // ( tok )
    create                            // And we need some data record for later use
      sm.states sm.hereEnclose        // Link here into sm.states list
      ,                               // Enclose nStr
    does>                             // Nothing, returning the pfa is default behaviour
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.addEvent ( nStr -- )

    sm.eExpand                        // Build <machineName>_e_<nStr>
    dup dup sm.prefix._               // ( tok tok _tok )
    swap sm.pushToken                 // ( tok _tok )
    sm.newEvent                       // Define a new event constant named tok
    sm.pushToken                      // ( tok )
    create                            // And we need some data record for later use
      sm.events sm.hereEnclose        // Link here into sm.events list
      ,                               // Enclose nStr
    does>                             // Nothing, returning the pfa is default behaviour
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : sm.addTransition ( nStr -- )

    sm.tExpand                        // Build <machineName>_t_<nStr>
    dup sm.prefix._                   // ( tok _tok )
    sm.pushToken                      // ( tok )
    create                            // We need some data record for later use
      sm.transitions sm.hereEnclose   // Link here into sm.transitions list
      ,                               // Enclose nStr
    does>                             // Nothing, returning the pfa is default behaviour
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : stateMachine ( -- )

    sm.clearStates                    // Clear state list
    sm.clearEvents                    // Clear event list
    sm.clearTransitions               // Clear transition list
    sm.getToken sm.machineName !      // Set current machine name to input token
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  : endStateMachine ( -- )


    // Some code should be generated here ...
  ;

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .


  : transition ( -- )

    sm.getToken dup sm.findState?       not if sm.addState       else drop endif  // Old state
    sm.getToken dup sm.findEvent?       not if sm.addEvent       else drop endif  // Event
    sm.getToken dup sm.findState?       not if sm.addState       else drop endif  // New state
    sm.getToken dup sm.findTransition?  not if sm.addTransition  else drop endif  // Transition function
  ;


}

// Test code

  stateMachine Bill
    transition Idle     Timer.fired     Waiting goWaiting
    transition Booted   Timer.fired     Waiting goWaiting
    transition Booted   ButtonPressed   Pressed goPressed
    transition Pressed  ButtonReleased  Waiting goWaiting
  endStateMachine



endModule



